package no.statnett.ecp;

import org.apache.qpid.amqp_1_0.jms.impl.ConnectionFactoryImpl;

import javax.jms.*;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.*;

public class AMQPConnector {

    private static final DateTimeFormatter tmsF = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS");

    private static final String CONNECTION_FACTORY_NAME = "myClientJmsFactory";
    public static final String JNDI_QUEUE_REFERENCE = "queue";
    private static final String JNDI_PROPERTY_FILE = "jndi.properties";
    private static final String JNDI_CONNECTION_FACTORY_PREFIX = "connectionfactory.";
    private static final String JNDI_DESTINATION_PREFIX = "destination.";
    private static final String JNDI_QUEUE_PREFIX = "queue.";


    public static Map<String, String> findBrokerConfig(String urlPathArg, String keystoreFile, String keystorePW) throws IOException {
        Map<String, String> configMap = new HashMap<>();
        if (urlPathArg.startsWith("amqp")) {
            configMap.put("url", urlPathArg);
        } else {
            String url = "";
            File f = new File(urlPathArg);
            if (f.isDirectory()) {
                File prop = new File(urlPathArg + "/" + "ecp.properties");
                if (prop.exists()) { // ECP-config
                    String host = findHost(urlPathArg);
                    List<String> lines = Files.readAllLines(prop.toPath());
                    Map<String, String> map = makeMap(lines);
                    String auth = getProp(map, "internalBroker.useAuthentication", "true", true);
                    if (auth.equals("true")) {
                        url += "amqps://" + getProp(map, "internalBroker.auth.user", "endpoint") + ":" + getProp(map, "internalBroker.auth.password", "password") + "@" + host + ":";
                        findKeystoreInfo(configMap, map);
                    } else {
                        url += "amqp://" + host + ":";
                    }
                    url += getProp(map, "internalBroker.port", "5672");
                    configMap.put("url", url);
                    configMap.put("ecp", "true");
                    String ecpLog = getProp(map, "logging.file.name", "/var/log/ecp-endpoint/ecp.log", false);
                    configMap.put("ecp-log", ecpLog);
                } else {
                    prop = new File(urlPathArg + "/" + "edx.properties");
                    if (prop.exists()) { // EDX-config
                        String host = findHost(urlPathArg);
                        List<String> lines = Files.readAllLines(prop.toPath());
                        Map<String, String> map = makeMap(lines);
                        String auth = getProp(map, "internalBroker.useAuthentication", "true", true);
                        if (auth.equals("true")) {
                            url += "amqps://" + getProp(map, "internalBroker.auth.user", "toolbox") + ":" + getProp(map, "internalBroker.auth.password", "password") + "@" + host + ":";
                            findKeystoreInfo(configMap, map);
                        } else {
                            url += "amqp://" + host + ":";
                        }
                        url += getProp(map, "internalBroker.amqp.port", "5672");
                        configMap.put("url", url);
                        configMap.put("edx", "true");
                        configMap.put("yml", urlPathArg + "/" + "edx.yml");
                    } else {
                        throw new RuntimeException("Did not find ecp.properties nor edx.properties in " + urlPathArg);
                    }
                }
            } else {
                throw new RuntimeException("Argument " + urlPathArg + " is not a directory, if you intend it to be a URL make sure it starts with 'amqp'");
            }
        }
        if (keystoreFile != null) {
            configMap.put("keystore-file", keystoreFile);
        }
        if (keystorePW != null) {
            configMap.put("keystore-pw", keystorePW);
        }
        if (configMap.get("keystore-file") != null) {
            File f = new File(configMap.get("keystore-file"));
            if (!f.exists()) {
                throw new RuntimeException("Cannot find the referred (directly or indirectly) keystore file:  " + configMap.get("keystore-file"));
            }
        }
        return configMap;
    }

    private static void makeJNDIProperties(boolean produce, Map<String, String> configMap, String queueName) throws IOException {
        FileWriter fw = new FileWriter(JNDI_PROPERTY_FILE, false);
        String ksfile = configMap.get("keystore-file");
        String kspass = configMap.get("keystore-pw");
        String url = configMap.get("url");
        boolean keystoreArgs = ksfile != null && kspass != null && url.contains("amqps");
        if (keystoreArgs) {
            int rndId = new Random().nextInt(10000);
            url += "?ssl=true&clientid=myclient" + rndId + "&remote-host=default&sync-publish=false&trust-store=" + ksfile + "&trust-store-password=" + kspass + "&key-store-password=" + kspass + "&key-store=" + ksfile;
        }
        configMap.put("url-jndi", url);
        fw.write(JNDI_CONNECTION_FACTORY_PREFIX + CONNECTION_FACTORY_NAME + " = " + url + "\n");
        if (produce)
            fw.write(JNDI_DESTINATION_PREFIX + JNDI_QUEUE_REFERENCE + " = " + queueName + "\n");
        else
            fw.write(JNDI_QUEUE_PREFIX + JNDI_QUEUE_REFERENCE + " = " + queueName + "\n");
        fw.close();
    }

    public static Context getContext(boolean produce, Map<String, String> config, String queueName) throws IOException, NamingException {
        makeJNDIProperties(produce, config, queueName);
        Hashtable env = new Hashtable();
        env.put(Context.INITIAL_CONTEXT_FACTORY, "org.apache.qpid.amqp_1_0.jms.jndi.PropertiesFileInitialContextFactory");
        env.put(Context.PROVIDER_URL, JNDI_PROPERTY_FILE);
        Context context = new InitialContext(env);
        return context;
    }

    public static Connection createSession(Context context) throws NamingException, JMSException {
        ConnectionFactoryImpl factory = (ConnectionFactoryImpl) context.lookup(CONNECTION_FACTORY_NAME);
        Connection connection = factory.createConnection();
        connection.start();
        return connection;
    }

    private static void findKeystoreInfo(Map<String, String> configMap, Map<String, String> map) {
        String keystore = getProp(map, "internalBroker.keystore.location", "${dataDirectory}/authKeystore.jks");
        int varPosStart = keystore.indexOf("${");
        int varPostStopp = keystore.indexOf("}");
        if (varPosStart > -1 && varPostStopp > -1) {
            String refPropName = keystore.substring(varPosStart + 2, varPostStopp);
            if (map.get(refPropName) == null) {
                throw new RuntimeException("Could not find the referenced property " + refPropName + " (referenced in " + keystore + " property)");
            }
            String substKeystore = keystore.substring(0, varPosStart) + map.get(refPropName) + keystore.substring(varPostStopp + 1);
            configMap.put("keystore-file", substKeystore);
        } else {
            configMap.put("keystore-file", keystore);
        }
        configMap.put("keystore-pw", getProp(map, "internalBroker.keystore.password", "password"));
    }

    private static String findHost(String urlPathArg) {
        try {
            return Files.readString(Path.of(urlPathArg + "/host")).trim();
        } catch (Throwable t) {
            return "localhost";
        }
    }

    private static String getProp(Map<String, String> map, String name, String defaultValue) {
        return getProp(map, name, defaultValue, false);
    }

    private static String getProp(Map<String, String> map, String name, String defaultValue, boolean toLowerCase) {
        if (map.get(name) != null) {
            return toLowerCase ? map.get(name).trim().toLowerCase() : map.get(name).trim();
        } else {
            return defaultValue;
        }
    }

    private static Map<String, String> makeMap(List<String> lines) {
        Map map = new HashMap();
        for (String line : lines) {
            if (!line.trim().startsWith("#") && line.contains("=") && line.split("=").length == 2) {
                map.put(line.split("=")[0].trim(), line.split("=")[1].trim());
            }
        }
        return map;
    }

    public static BytesMessage buildBytesMessage(final Session session, final byte[] message, String receiverCode, String messageType) throws JMSException {
        final BytesMessage bytesMessage = session.createBytesMessage();
        bytesMessage.writeBytes(message);
        setMessageProperties(bytesMessage, receiverCode, messageType);
        return bytesMessage;
    }

    private static void setMessageProperties(final Message message, String receiverCode, String messageType) throws JMSException {
        message.setStringProperty("receiverCode", receiverCode);
        message.setStringProperty("messageType", messageType);
        message.setStringProperty("businessType", messageType);
        message.setStringProperty("senderApplication", "AMQPConnector");
        // More headers could be specified, but are not mandatory - see EDX User Guide chapter 5.
    }

    public static String buildSelector(String messagetType, String TTL, String senderCode, String senderBA, String receiverCode, String internalType) {
        String selector = null;
        if (!messagetType.equals("ALL")) {
            selector = "(messageType = '" + messagetType + "' OR businessType = '" + messagetType + "')";
        }
        if (TTL != null && Long.parseLong(TTL) > 0) {
            String newSelector = "JMSTimestamp < " + (System.currentTimeMillis() - (Long.parseLong(TTL) * 1000L));
            selector = appendSelector(selector, newSelector);
        }
        if (senderCode != null) {
            String newSelector = "senderCode = '" + senderCode + "'";
            selector = appendSelector(selector, newSelector);
        }
        if (senderBA != null) {
            String newSelector = "baMessageType = '" + senderBA + "'";
            selector = appendSelector(selector, newSelector);
        }
        if (receiverCode != null) {
            String newSelector = "receiverCode = '" + receiverCode + "'";
            selector = appendSelector(selector, newSelector);
        }
        if (internalType != null) {
            String newSelector = "internalType = '" + internalType + "'";
            selector = appendSelector(selector, newSelector);
        }

        return selector;
    }

    private static String appendSelector(String selector, String TTLSelect) {
        selector = (selector == null ? TTLSelect : selector + " AND " + TTLSelect);
        return selector;
    }

    public static void logMessage(boolean dryrun, Message msg, String queue) throws JMSException {
        LocalDateTime generatedLDT = tmsLDT(msg.getJMSTimestamp());
        String foundMT = msg.getStringProperty("messageType");
        if (foundMT == null)
            foundMT = msg.getStringProperty("businessType");
        Long ageSeconds = null;
        if (generatedLDT != null) {
            LocalDateTime now = LocalDateTime.now(ZoneOffset.UTC);
            ageSeconds = ChronoUnit.SECONDS.between(generatedLDT, now);
        }
        String recevierCode = msg.getStringProperty("receiverCode");
        String sender = msg.getStringProperty("senderCode");
        String senderBA = msg.getStringProperty("senderBA");

        long hours = ageSeconds / 3600;
        long mins = (ageSeconds % 3600) == 0 ? 0 : (ageSeconds % 3600) / 60;
        String hourStr = hours + "h " + String.format("%02d", mins) + "m";
        String status = dryrun ? "Browsed" : "Consumed";
        System.out.println(LocalDateTime.now().format(tmsF) + "  DEBUG [queue=" + String.format("%-55s", queue) + "]      " + status + ": BTMT=" + foundMT + ", Age=" + ageSeconds + " sec (" + hourStr + "), Sender=" + sender + ", Receiver=" + recevierCode + ", SenderBA=" + senderBA);
    }

    protected static LocalDateTime tmsLDT(Long epoch) {
        if (epoch == null)
            return null;
        return Instant.ofEpochMilli(epoch).atZone(ZoneOffset.UTC).toLocalDateTime();
    }

    public static List<String> findQueuenames(String arg) throws IOException {
        if (arg.contains("yml")) {
            return findEDXQueues(arg);
        } else if (arg.contains("ecp.properties")) {
            return findECPQueues();
        } else {
            return Arrays.asList(arg.split(","));
        }
    }

    private static List<String> findECPQueues() {
        List<String> queues = new ArrayList<>();
        queues.add("ActiveMQ.DLQ");
        queues.add("ecp.endpoint.archive");
        queues.add("ecp.endpoint.download");
        queues.add("ecp.endpoint.handler.receive");
        queues.add("ecp.endpoint.handler.send");
        queues.add("ecp.endpoint.outbox");
        queues.add("ecp.endpoint.outbox.reply");
        queues.add("ecp.endpoint.internal.outbox");
        queues.add("ecp.endpoint.inbox");
        queues.add("ecp.endpoint.send.event");
        //TODO Mangler upload-køer
        return queues;
    }

    private static List<String> findEDXQueues(String edxYML) throws IOException {
        List<String> queues = new ArrayList<>();
        for (String line : Files.readAllLines(Path.of(edxYML))) {
            if (line.trim().startsWith("#")) {
                continue;
            }
            queues.add(findValueInYMLConfig(line, "queueName:"));
            queues.add(findValueInYMLConfig(line, "replyQueueName:"));
        }
        for (String line : Files.readAllLines(Path.of(edxYML))) {
            if (line.trim().startsWith("#")) {
                continue;
            }
            String code = findValueInYMLConfig(line, "code:");
            if (code != null) {
                queues.add("edx.internal.delivery." + code);
                queues.add("edx.internal.replyDelivery." + code);
            }
            if (line.indexOf("routes:") > -1) {
                break;
            }
        }
        // Important fixed queues for EDX
        queues.add("ActiveMQ.DLQ");
        queues.add("edx.internal.receiveProcess-process");
        queues.add("edx.internal.sendProcess-delivery-send");
        queues.add("edx.internal.sendProcess-send");
        queues.removeAll(Collections.singleton(null));
        return queues;
    }

    private static String findValueInYMLConfig(String line, String searchTerm) {
        int searchTermStartPos = line.indexOf(searchTerm);
        int searchTermEndPos = searchTermStartPos + searchTerm.length();

        if (searchTermStartPos > -1) {
            int endPos = Integer.MAX_VALUE;
            int commaEndPos = line.indexOf(",", searchTermEndPos);
            if (commaEndPos > -1) {
                endPos = commaEndPos;
            }
            int endBracketPos = line.indexOf("}", searchTermEndPos);
            if (endBracketPos > -1 && endBracketPos < endPos) {
                endPos = endBracketPos;
            }
            if (endPos == Integer.MAX_VALUE) { // line ends without , or }
                String queueName = line.substring(searchTermEndPos).trim();
                return queueName;
            } else {
                String queueName = line.substring(searchTermEndPos, endPos).trim();
                return queueName;
            }
        }
        return null;
    }

    public static Map<String, String> findFilterArgs(String arg) {
        String[] filterArgs = arg.split(",");
        Map<String, String> filterMap = new HashMap<>();
        filterMap.put("BTMT", filterArgs[0]);
        for (int i = 1; i < filterArgs.length; i++) {
            if (filterArgs[i].startsWith("TTL-")) {
                filterMap.put("TTL", filterArgs[i].substring(4));
            } else if (filterArgs[i].startsWith("SEN-")) {
                filterMap.put("SEN", filterArgs[i].substring(4));
            } else if (filterArgs[i].startsWith("SER-")) {
                filterMap.put("SER", filterArgs[i].substring(4));
            } else if (filterArgs[i].startsWith("BUA-")) {
                filterMap.put("BUA", filterArgs[i].substring(4));
            } else if (filterArgs[i].startsWith("REC-")) {
                filterMap.put("REC", filterArgs[i].substring(4));
            } else if (filterArgs[i].startsWith("IT-")) {
                filterMap.put("IT", filterArgs[i].substring(3));
            }
        }
        return filterMap;
    }


}
