/*
 * Copyright (C) Red Hat, Inc.
 * http://www.redhat.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package no.statnett.ecp;

import javax.jms.Queue;
import javax.jms.*;
import javax.naming.Context;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;

public class QueueCleaner {
    private static final DateTimeFormatter tmsF = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS");
    private static final DateTimeFormatter tms2F = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");

    private static String VERSION = "v1.6.0";

    private static final Boolean NON_TRANSACTED = false;
    private static final String DESTINATION_NAME = "queue/browse";

    private static Map<String, Integer> antallMeldingerPrMeldingstype;
    private static Map<String, Integer> antallMeldingerPrSender;
    private static Map<String, Integer> antallMeldingerPrSenderBA;
    private static Map<String, Integer> antallMeldingerPrReceiver;
    private static Map<String, Integer> antallMeldingerPrService;
    private static Map<String, Integer> antallMeldingerPrTidsepoke;
    private static Map<String, Integer> antallMeldingerPrInternalType;
    private static Map<String, Integer> antallMeldingerPrQueue = new TreeMap<>();

    private static String filterTTL;
    private static String filterBTMT;
    private static String filterSender;
    private static String filterSenderBA;
    private static String filterReceiver;
    private static String filterService;
    private static String filterInternalType;
    private static boolean verbose = false; // default
    private static boolean dryrun = false; // default
    private static int maxConsume = Integer.MAX_VALUE; // default
    private static int consumeTimeout = 100;
    private static int consumeCounter = 0;
    private static int browserCounter = 0;
    private static File outputFile;
    private static String status = "OK";
    private static List<String> queues = new ArrayList<>();

    public static void main(String args[]) throws IOException {

        args = parseOptions(args);
        if (args.length < 3) {
            exit(true);
        } else if (args[0].startsWith("amqps") && args.length < 5) {
            System.err.println("The URL indicates AMQPS, but no truststore or password file is given - see usage.");
            exit(false);
        }
        findFilterArgs(args);

        Map<String, String> config = AMQPConnector.findBrokerConfig(args[0], args.length > 3 ? args[3] : null, args.length > 4 ? args[4] : null);

        findQueues(config, args[1]);

        Connection connection = null;
        boolean firstRound = true;
        for (String queueName : queues) {
            try {
                Context context = AMQPConnector.getContext(false, config, queueName);
                connection = AMQPConnector.createSession(context);
                Session session = connection.createSession(NON_TRANSACTED, Session.AUTO_ACKNOWLEDGE);


                if (verbose && firstRound) {
                    System.out.println("Starting QueueCleaner " + VERSION + ", have deduced AMQP-URL: " + config.get("url-jndi"));
                    firstRound = false;
                }


                Queue queue = (Queue) context.lookup(DESTINATION_NAME);

                antallMeldingerPrQueue.merge(queueName, 0, Integer::sum);

                antallMeldingerPrMeldingstype = new TreeMap<>();
                antallMeldingerPrTidsepoke = new TreeMap<>();
                antallMeldingerPrSender = new TreeMap<>();
                antallMeldingerPrSenderBA = new TreeMap<>();
                antallMeldingerPrService = new TreeMap<>();
                antallMeldingerPrReceiver = new TreeMap<>();
                antallMeldingerPrInternalType = new TreeMap<>();


                if (dryrun) {
                    System.out.println(info(queueName) + "Browsing with filter BTMT=" + filterBTMT + ", TTL=" + filterTTL + ", SEN=" + filterSender + ", REC=" + filterReceiver + ", BAU=" + filterSenderBA + ", IT=" + filterInternalType);
                    QueueBrowser browser = session.createBrowser(queue, AMQPConnector.buildSelector(filterBTMT, filterTTL, filterSender, filterSenderBA, filterReceiver, filterInternalType));
                    Enumeration<?> enum1 = browser.getEnumeration();
                    while (enum1.hasMoreElements()) {
                        Object melding = enum1.nextElement();
                        logMessage((Message) melding, queueName);
                        browserCounter++;
                    }
                    browser.close();
                } else {
                    System.out.println(info(queueName) + "Consuming with filter BTMT=" + filterBTMT + ", TTL=" + filterTTL + ", SEN=" + filterSender + ", REC=" + filterReceiver + ", BAU=" + filterSenderBA + ", IT=" + filterInternalType);
                    MessageConsumer consumer = session.createConsumer(queue, AMQPConnector.buildSelector(filterBTMT, filterTTL, filterSender, filterSenderBA, filterReceiver, filterInternalType));
                    while (true) {
                        Message message;
                        if (consumeTimeout == 0 && queues.size() == 1) { // 'tails' on a specific queue - will block until a message arrives
                            message = consumer.receive();
                        } else { // will not block, but return null if no message is available
                            message = consumer.receive(100);
                        }
                        if (message == null)
                            break;
                        message.acknowledge();
                        logMessage(message, queueName);
                        consumeCounter++;
                        if (consumeCounter >= maxConsume)
                            break;
                    }
                }
                logSummary(queueName);
            } catch (Exception t) {
                System.err.println(error(queueName) + "Error occurred: " + t);
                status = t.getMessage().replaceAll("\"", "");
            } finally {
                if (connection != null) {
                    try {
                        connection.close();
                    } catch (JMSException e) {
                        System.err.println(error(queueName) + "Error occurred while close JMS-connection:" + e);
                    }
                }
            }
        }
        logTotalSummary();
    }


    private static void logTotalSummary() throws IOException {
        if (outputFile != null && outputFile.exists()) {
            StringBuilder sb = new StringBuilder();
            sb.append("{ \"statustime\": \"" + LocalDateTime.now().format(tms2F) + "\", ");
            sb.append("\"status\": \"" + status + "\", ");
            sb.append("\"queues\": [");
            List<String> queueStrList = new ArrayList<>();
            for (String queueName : antallMeldingerPrQueue.keySet()) {
                queueStrList.add("{ \"queue\": \"" + queueName + "\", \"no_of_messages\": " + antallMeldingerPrQueue.get(queueName) + " }");
            }
            sb.append(String.join(",", queueStrList));
            sb.append("]}\n");
            Files.writeString(outputFile.toPath(), sb.toString(), StandardOpenOption.APPEND);
        }
        if (dryrun) {
            System.out.println(info("Summary") + String.format("%5s", browserCounter) + " messages was browsed in total");
        }
        System.out.println(info("Summary") + String.format("%5s", consumeCounter) + " messages was consumed in total");
    }

    private static String[] parseOptions(String[] initialArgs) {
        List<String> args = new ArrayList<>();
        boolean nextArgIsOutputfile = false;
        for (int i = 0; i < initialArgs.length; i++) {
            if (nextArgIsOutputfile) {
                outputFile = new File(initialArgs[i]);
                try {
                    if (outputFile.exists()) {
                        Files.writeString(outputFile.toPath(), "", StandardOpenOption.APPEND);
                    } else {
                        Files.writeString(outputFile.toPath(), "");
                    }
                    nextArgIsOutputfile = false;
                } catch (IOException e) {
                    System.err.println("Cannot write " + initialArgs[i] + " (exception: " + e.getMessage() + ")");
                    exit(false);
                }
            } else if (initialArgs[i].startsWith("-v")) {
                verbose = true;
            } else if (initialArgs[i].startsWith("-d")) {
                dryrun = true;
            } else if (initialArgs[i].startsWith("-m")) {
                maxConsume = Integer.parseInt(initialArgs[i].substring(2));
            } else if (initialArgs[i].startsWith("-f")) {
                consumeTimeout = 0;
            } else if (initialArgs[i].startsWith("-o")) {
                nextArgIsOutputfile = true;
            } else {
                args.add(initialArgs[i]);
            }
        }
        return args.toArray(new String[]{});
    }


    private static void exit(boolean showUsage) {
        if (showUsage)
            usage();
        else
            System.err.println("Tip: If you want to run without SSL, check that interalBroker.useAuthentication=false in ecp-/edx.properties or specify AMQP-URL directly. You can try to add user/pass in the URL.");
        System.exit(1);
    }

    private static void findQueues(Map<String, String> config, String arg) throws IOException {
        if (arg.contains("ALL")) {
            if (config.get("ecp") == null && config.get("edx") == null) {
                System.err.println("Cannot find ALL queues for ECP/EDX unless path to ecp.properties/edx.yml is specified - see usage");
                exit(false);
            } else if (config.get("ecp") != null) {
                queues = findECPQueues(findUploadQueues(config.get("ecp-log")));
            } else if (config.get("edx") != null) {
                if (config.get("yml") != null)
                    queues = findEDXQueues(config.get("yml"));
                else
                    queues = findEDXQueues(arg);
            }
        } else if (arg.contains("ecp.properties")) {
            queues = findECPQueues(findUploadQueues(config.get("ecp-log")));
        } else {
            queues = Arrays.asList(arg.split(","));
        }
    }

    private static String[] findUploadQueues(String ecpLogPath) {
        Set<String> uploadQueues = new HashSet<>();
        final AtomicInteger counter = new AtomicInteger(0);
        try (Stream<String> stream = Files.lines(Paths.get(ecpLogPath), StandardCharsets.UTF_8)) {
            // Linje vi leter etter:
            // 2023-11-13 13:17:29.808  INFO 3853781 --- [Camel (camel-1) thread #25 - JmsConsumer[ecp.endpoint.outbox]] EcpAuditLog                              : ComponentCode=50V-SN-DKBSP-AT2, ComponentType=ECP Endpoint, RecordID=32a95be7-f7c6-460d-8d09-489a4f7d9c7f, ID=MSG-11, Category=MSG, Text=Message [messageID=96a99caa-8793-403d-bc78-ba42463d7ffc,internalType=STANDARD_MESSAGE,senderCode=50V-SN-DKBSP-AT2,receiverCode=50V-HAFSE-CO-ATU,businessType=NBM-MFRREAM-CIM-PTA47-MTA41-ACK,baMessageID=ERRP-20231113141728417-0-b2,senderApplication=Fifty,generated=2023-11-13T14:17:29.467+01:00[Europe/Brussels]] has been uploaded to 50V000000000112U., Direction=OUT, User=SYSTEM, Severity=0
            // Hvis skal finne stringen som står etter "uploaded to " og før "." og deretter legge den til uploadQueues
            stream.forEach(line -> {
                counter.incrementAndGet();
                if (line.contains("ID=MSG-11") && line.contains("uploaded to ")) {
                    int startPos = line.indexOf("uploaded to ") + 12;
                    int endPos = line.indexOf(".", startPos);
                    String queueName = line.substring(startPos, endPos);
                    uploadQueues.add(queueName);
                }
            });
        } catch (IOException e) {
            e.printStackTrace();
        }
        System.out.println("Search through " + counter.get() + " rows in " + ecpLogPath + " (referenced in ecp.properties) and found " + uploadQueues.size() + " upload queues");
        return uploadQueues.toArray(new String[]{});
    }

    private static List<String> findECPQueues(String[] brokers) {
        List<String> queues = new ArrayList<>();
        queues.add("ActiveMQ.DLQ");
        queues.add("ecp.endpoint.archive");
        queues.add("ecp.endpoint.download");
        queues.add("ecp.endpoint.handler.receive");
        queues.add("ecp.endpoint.handler.send");
        queues.add("ecp.endpoint.outbox");
        queues.add("ecp.endpoint.outbox.reply");
        queues.add("ecp.endpoint.internal.outbox");
        queues.add("ecp.endpoint.inbox");
        queues.add("ecp.endpoint.send.event");
        for (String broker : brokers) {
            queues.add("ecp.endpoint.upload." + broker);
        }
        return queues;
    }


    private static List<String> findEDXQueues(String edxYML) throws IOException {
        List<String> queues = new ArrayList<>();
        for (String line : Files.readAllLines(Path.of(edxYML))) {
            if (line.trim().startsWith("#")) {
                continue;
            }
            queues.add(findValueInYMLConfig(line, "queueName:"));
            queues.add(findValueInYMLConfig(line, "replyQueueName:"));
        }
        for (String line : Files.readAllLines(Path.of(edxYML))) {
            if (line.trim().startsWith("#")) {
                continue;
            }
            String code = findValueInYMLConfig(line, "code:");
            if (code != null) {
                queues.add("edx.internal.delivery." + code);
                queues.add("edx.internal.replyDelivery." + code);
            }
            if (line.indexOf("routes:") > -1) {
                break;
            }
        }
        // Important fixed queues for EDX
        queues.add("ActiveMQ.DLQ");
        queues.add("edx.internal.receiveProcess-process");
        queues.add("edx.internal.sendProcess-delivery-send");
        queues.add("edx.internal.sendProcess-send");
        queues.removeAll(Collections.singleton(null));
        return queues;
    }

    private static String findValueInYMLConfig(String line, String searchTerm) {
        int searchTermStartPos = line.indexOf(searchTerm);
        int searchTermEndPos = searchTermStartPos + searchTerm.length();

        if (searchTermStartPos > -1) {
            int endPos = Integer.MAX_VALUE;
            int commaEndPos = line.indexOf(",", searchTermEndPos);
            if (commaEndPos > -1) {
                endPos = commaEndPos;
            }
            int endBracketPos = line.indexOf("}", searchTermEndPos);
            if (endBracketPos > -1 && endBracketPos < endPos) {
                endPos = endBracketPos;
            }
            if (endPos == Integer.MAX_VALUE) { // line ends without , or }
                String queueName = line.substring(searchTermEndPos).trim();
                return queueName;
            } else {
                String queueName = line.substring(searchTermEndPos, endPos).trim();
                return queueName;
            }
        }
        return null;
    }

    public static void findFilterArgs(String[] args) {
        String[] filterArgs = args[2].split(",");
        filterBTMT = filterArgs[0];
        for (int i = 1; i < filterArgs.length; i++) {
            if (filterArgs[i].startsWith("TTL-")) {
                filterTTL = filterArgs[i].substring(4);
            } else if (filterArgs[i].startsWith("SEN-")) {
                filterSender = filterArgs[i].substring(4);
            } else if (filterArgs[i].startsWith("SER-")) {
                filterService = filterArgs[i].substring(4);
            } else if (filterArgs[i].startsWith("BUA-")) {
                filterSenderBA = filterArgs[i].substring(4);
            } else if (filterArgs[i].startsWith("REC-")) {
                filterReceiver = filterArgs[i].substring(4);
            } else if (filterArgs[i].startsWith("IT-")) {
                filterInternalType = filterArgs[i].substring(3);
            }
        }
    }


    private static void logSummary(String queueName) {
        if (filterBTMT.equals("ALL")) {
            for (Map.Entry<String, Integer> entry : antallMeldingerPrMeldingstype.entrySet()) {
                System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages of Message/BusinessType " + entry.getKey());
            }
        }
        for (Map.Entry<String, Integer> entry : antallMeldingerPrTidsepoke.entrySet()) {
            System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages generated during the past " + entry.getKey());
        }
        if (filterSender == null) {
            for (Map.Entry<String, Integer> entry : antallMeldingerPrSender.entrySet()) {
                System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages of Sender " + entry.getKey());
            }
        }
        if (filterService == null) {
            for (Map.Entry<String, Integer> entry : antallMeldingerPrService.entrySet()) {
                System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages of Service " + entry.getKey());
            }
        }
        if (filterSenderBA == null) {
            for (Map.Entry<String, Integer> entry : antallMeldingerPrSenderBA.entrySet()) {
                System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages of SenderBA " + entry.getKey());
            }
        }
        if (filterReceiver == null) {
            for (Map.Entry<String, Integer> entry : antallMeldingerPrReceiver.entrySet()) {
                System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages of Receiver " + entry.getKey());
            }
        }
        if (filterInternalType == null) {
            for (Map.Entry<String, Integer> entry : antallMeldingerPrInternalType.entrySet()) {
                System.out.println(info(queueName) + String.format("%5s", entry.getValue()) + " messages of InternalType " + entry.getKey());
            }
        }

        System.out.println(info(queueName) + String.format("%5s", antallMeldingerPrTidsepoke.values().stream().mapToInt(i -> i).sum()) + " in total");
    }

    private static String error(String queue) {
        return LocalDateTime.now().format(tmsF) + "  ERROR [queue=" + String.format("%-55s", queue) + "] ";
    }

    private static String info(String queue) {
        return LocalDateTime.now().format(tmsF) + "   INFO [queue=" + String.format("%-55s", queue) + "] ";
    }

    private static String debug(String queue) {
        return LocalDateTime.now().format(tmsF) + "  DEBUG [queue=" + String.format("%-55s", queue) + "] ";
    }


    private static void logMessage(Message msg, String queueName) throws JMSException {
        antallMeldingerPrQueue.merge(queueName, 1, Integer::sum);
        String msgBusinessType = findBusinessMessageType(msg);
        if (msgBusinessType != null) {
            antallMeldingerPrMeldingstype.merge(msgBusinessType, 1, Integer::sum);
        }
        LocalDateTime generatedLDT = AMQPConnector.tmsLDT(msg.getJMSTimestamp());
        String foundMT = msg.getStringProperty("messageType");
        if (foundMT == null)
            foundMT = msg.getStringProperty("businessType");
        Long ageSeconds = null;
        if (generatedLDT != null) {
            LocalDateTime now = LocalDateTime.now(ZoneOffset.UTC);
            ageSeconds = ChronoUnit.SECONDS.between(generatedLDT, now);
            antallMeldingerPrTidsepoke.merge(tidsepoke(ageSeconds), 1, Integer::sum);
        }
        String recevierCode = msg.getStringProperty("receiverCode");
        if (recevierCode != null) {
            antallMeldingerPrReceiver.merge(recevierCode, 1, Integer::sum);
        }
        String sender = msg.getStringProperty("senderCode");
        if (sender != null) {
            antallMeldingerPrSender.merge(sender, 1, Integer::sum);
        }
        String senderBA = msg.getStringProperty("senderBA");
        if (senderBA != null) {
            antallMeldingerPrSenderBA.merge(senderBA, 1, Integer::sum);
        }
        String internalType = msg.getStringProperty("internalType");
        if (internalType != null) {
            antallMeldingerPrInternalType.merge(internalType, 1, Integer::sum);
        }

        String payloadSize = "N/A";
        if (msg instanceof TextMessage) {
            payloadSize = "" + ((TextMessage) msg).getText().length();
        } else if (msg instanceof BytesMessage) {
            payloadSize = ((BytesMessage) msg).getBodyLength() + " bytes";
        }

        long hours = ageSeconds / 3600;
        long mins = (ageSeconds % 3600) == 0 ? 0 : (ageSeconds % 3600) / 60;
        String hourStr = hours + "h " + String.format("%02d", mins) + "m";
        String status = dryrun ? "Browsed" : "Consumed";
        System.out.println(debug(queueName) + "      " + status + ": BTMT=" + foundMT + ", Age=" + ageSeconds + " sec (" + hourStr + "), Sender=" + sender + ", Receiver=" + recevierCode + ", SenderBA=" + senderBA + ", InternalType=" + internalType + ", PayloadSize=" + payloadSize);
    }

    private static String findBusinessMessageType(Message msg) throws JMSException {
        String msgBusinessType = msg.getStringProperty("businessType");
        if (msgBusinessType == null) {
            msgBusinessType = msg.getStringProperty("messageType");
        }
        return msgBusinessType;
    }


    private static void usage() {
        System.out.println("QueueCleaner " + VERSION + " is a tool to browse or consume messages from all queues on ECP or EDX - it can be used to monitor or to");
        System.out.println("maintain the queues on brokers of ECP and EDX. It can be run locally or remotely. It can filter for certain messages, and");
        System.out.println("by so doing, be able to clean up certain messages that you want to get rid of.\n\n");
        System.out.println("Usage  : java -jar QueueCleaner.jar [OPTION] BROKER QUEUES FILTER [KEYSTORE]\n");
        System.out.println("\n\nThe options and arguments:");
        System.out.println(" OPTION        : ");
        System.out.println("                 -d                 : dryrun/browse - will not consume messages. Maximum 400 msg can be browsed");
        System.out.println("                 -m<number>         : restrict <number> of messages to be consumed. Default is to consume all messages");
        System.out.println("                 -o <filename>      : will append a one-liner summary in JSON-format to the <filename>");
        System.out.println("                 -v                 : prints the deduced URL to console - for troubleshooting");
        System.out.println("                 -f                 : 'tails' on a queue - this option will only work if you specify a single queue below");
        System.out.println(" BROKER        : <URL|PATH> - either specify a URL to a broker or a path for the ECP- or EDX-configuration");
        System.out.println("                 URL                : Specify url starting with either 'amqp' or 'amqps' + keystore-file/password args at the end");
        System.out.println("                 PATH               : Specify either the conf-path for ECP or EDX (ex: /etc/ecp-endpoint or /etc/edx-toolbox)");
        System.out.println(" QUEUES        :                    : Number of different options to specify the queues to scan: ");
        System.out.println("                 (<queue-name>[,<queue-name>]*)    : will scan these queues");
        System.out.println("                 ALL                               : will scan ALL queues, but this requires PATH in previous arg");
        System.out.println("                 <edx.yml>                         : will scan ALL queues from edx.yml");
        System.out.println("                 <ecp.properties>                  : will scan ALL queues from ecp.properties (currently: only a fixed list)");
        System.out.println(" FILTER        : (<messageType>|ALL)[,TTL-<sec>][,SEN-<senderCode>][,REC-<receiverCode>][,BAU-<businessApp>] - see below for expl.");
        System.out.println("                 <messageType>|ALL  : Mandatory filter - will match both on businessType and messageType. Keyword 'ALL' matches all msg");
        System.out.println("                 TTL-<sec>          : Optional filter to specify max <sec> TTL/age for the message - if omitted it matches all msg");
        System.out.println("                 SEN-<senderCode>   : Optional filter to specify <senderCode> for the message - if omitted it matches all msg");
        System.out.println("                 REC-<receiverCode> : Optional filter to specify <receiverCode> for the message - if omitted it matches all msg");
        System.out.println("                 BAU-<businessApp>  : Optional filter to specify <BA> for the message - if omitted it matches all msg");
        System.out.println("                 IT-<internalType>  : Optional filter to specify STANDARD_MESSAGE, DELIVERY_ACKNOWLEDGEMENT, RECEIVE_ACKNOWLEDGEMENT, TRACING_MESSAGE");
        System.out.println(" KEYSTORE        <keystore-file> <keystore-password> - these two parameters are only necessary if you specify 'amqps' as URL argument. ");
        System.out.println("                                                       If they are specified they will override keystore params found in PATH-config");

        System.out.println("\nExample 1:  Consume all messages from standard ActiveMQ DLQ:");
        System.out.println("\tjava -jar QueueCleaner.jar amqp://localhost:5672 ActiveMQ.DLQ ALL");

        System.out.println("\nExample 2:  Consume older than 1h messages on edx.endpoint.reply (case: typically leftover msg that BA is not consuming):");
        System.out.println("\tjava -jar QueueCleaner.jar amqp://localhost:5672 edx.endpoint.reply ALL,TTL-3600");

        System.out.println("\nExample 3:  Consume older than 1h messages on ecp.endpoint.inbox and ecp.endpoint.send.event (case: msg stuck on-route from ECP to EDX):");
        System.out.println("\tjava -jar QueueCleaner.jar amqp://localhost:5672 ecp.endpoint.inbox,ecp.endpoint.send.event ALL,TTL-3600");

        System.out.println("\nExample 4:  Consume max 40 message from DLQ (use SSL) (case: consume 1d old AOL-msg from DLQ sent from ECO-2 endpoint):");
        System.out.println("\tjava QueueCleaner.jar -m40 amqps://toolbox:password@localhost:5672 ActiveMQ.DLQ NBM-CIM-PTZ12-MTZ35,TTL-86400,SEN-50V-NBM-E2HT-ATP authKeystore.jks password");

        System.out.println("\nExample 5:  Browse first 400 messages from standard ActiveMQ DLQ:");
        System.out.println("\tjava -jar QueueCleaner.jar -d amqp://localhost:5672 ActiveMQ.DLQ ALL");

        System.out.println("\nExample 6:  Browse first 400 messages from all queues on EDX:");
        System.out.println("\tjava -jar QueueCleaner.jar -d /etc/edx-toolbox ALL ALL");

        System.out.println("\nExample 7:  Consume all 1h old messages from all standard queues on ECP. Search each queue for up till 5000ms.");
        System.out.println("\tjava -jar QueueCleaner.jar /etc/ecp-endpoint ALL ALL,TTL-3600");

        System.out.println("\nExample 8:  Browse first 400 10-min old messages from all queues found in edx.yml:");
        System.out.println("\tjava -jar QueueCleaner.jar -d amqp://localhost:5672 /etc/edx-toolbox/edx.yml ALL,TTL-600");

        System.out.println("\nExample 9:  Browse first 400 10-min old messages from all queues found in edx.yml and print summary to json-file:");
        System.out.println("\tjava -jar QueueCleaner.jar -d amqp://localhost:5672 /etc/edx-toolbox/edx.yml ALL,TTL-600 -o /var/log/edx-toolbox/qc.json");

        System.out.println("\nExample 10: Consume continuously on a single queue - will block until a message arrives:");
        System.out.println("\tjava -jar QueueCleaner.jar -f amqp://localhost:5672 ActiveMQ.DLQ ALL");

        System.out.println("\nTo log with UTC time, start command with 'java -Duser.timezone=GMT -jar QueueCleaner.jar....etc'");
        System.out.println("The keystore must contain at the very least a root-CA which QueueCleaner can trust when connecting to the broker");
        System.out.println("If the QueueCleaner fails, take a look at the jndi.properties file that it makes and examine the URL it uses");
    }

    private static String tidsepoke(Long ageSeconds) {
        if (ageSeconds < 60) {
            return "minute";
        } else if (ageSeconds < 3600) {
            return "1-60 minutes";
        } else if (ageSeconds < 6 * 3600) {
            return "1-6 hours";
        } else if (ageSeconds < 86400) {
            return "6-24 hours";
        } else if (ageSeconds < 86400 * 7) {
            return "24-168 hours";
        } else if (ageSeconds < 86400 * 14) {
            return "168-336 hours";
        } else return "336 hours or earlier";
    }
}