package no.statnett.ecp.burst;

import no.statnett.ecp.AMQPConnector;

import javax.jms.*;
import javax.naming.Context;
import java.io.IOException;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;

public class BurstConsumerSteadyProducer {

  private static String VERSION = "v0.9.2";

  public static void main(String args[]) throws IOException {
    if (args.length < 4 || args.length > 6) {
      usage();
      System.exit(1);
    }
    Integer maxConsume = parseMaxConsume(args);
    String brokerArg = args[1];
    List<String> Qnames = AMQPConnector.findQueuenames(args[2]);
    Map<String, String> filter = AMQPConnector.findFilterArgs(args[3]);
    String keystoreFile = null; // specify file if using amqps in urlPathArg, example if connecting to ECP: "/var/lib/ecp-endpoint/authKeystore.jks"
    String keystorePW = null; // specify file if using amqps in urlPathArg, example: "password"
    if (args.length > 4) {
      keystoreFile = args[4];
      keystorePW = args[5];
    }

    try {
      int consumeCounter = 0;
      Map<String, String> config = AMQPConnector.findBrokerConfig(brokerArg, keystoreFile, keystorePW);
      for (String Qname : Qnames) {
        System.out.println("Scanning " + Qname);
        Context context = AMQPConnector.getContext(false, config, Qname);
        Connection connection = AMQPConnector.createSession(context);
        Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
        String selector = AMQPConnector.buildSelector(filter.get("BTMT"), filter.get("TTL"), filter.get("SEN"), filter.get("BAU"), filter.get("REC"), filter.get("IT"));
        if (maxConsume != null && maxConsume == 0) {
          QueueBrowser browser = session.createBrowser((Queue) context.lookup(AMQPConnector.JNDI_QUEUE_REFERENCE), selector);
          Enumeration<?> enum1 = browser.getEnumeration();
          while (enum1.hasMoreElements()) {
            Object melding = enum1.nextElement();
            AMQPConnector.logMessage(true, (Message) melding, Qname);
          }
          browser.close();
        } else {
          MessageConsumer consumer = session.createConsumer((Queue) context.lookup(AMQPConnector.JNDI_QUEUE_REFERENCE), selector);
          while (true) {
            Message message = consumer.receive(1000);
            if (message == null)
              break;
            message.acknowledge();
            AMQPConnector.logMessage(false, message, Qname);
            if (maxConsume != null && consumeCounter >= maxConsume) {
              break;
            }
          }
          consumer.close();
        }
        session.close();
      }
      System.out.println("Scanned " + Qnames.size() + " queues");
    } catch (Throwable t) {
      System.err.println("Error receiving message" + t);
    }
  }

  private static Integer parseMaxConsume(String[] args) {
    if (args[0].equals("ALL")) {
      return null;
    } else if (args[0].equals("NONE")) {
      return 0;
    } else {
      return Integer.parseInt(args[0]);
    }
  }

  private static void usage() {
    System.out.println("ConsumerExample " + VERSION + " is a demonstration on how to consume from ECP/EDX over AMQP(S). It can be used like this:\n\n");
    System.out.println("Usage  : java -jar ConsumerExample.jar CONSUME BROKER QUEUES FILTER [KEYSTORE]\n");
    System.out.println("\n\nThe options and arguments:");
    System.out.println(" CONSUME       : ALL|<INT>|NONE");
    System.out.println("                 ALL                : consume all messages matching FILTER");
    System.out.println("                 <INT>              : consume <INT> number of messages matching FILTER. 0 is eqiuvalent with NONE");
    System.out.println("                 NONE               : browse (read-only) max 400 messages matching FILTER");
    System.out.println(" BROKER        : <URL|PATH> - either specify a URL to a broker or a path for the ECP- or EDX-configuration");
    System.out.println("                 URL                : Specify url starting with either 'amqp' or 'amqps' (the latter requires extra args - see below");
    System.out.println("                 PATH               : Specify either the conf-path for ECP or EDX (ex: /etc/ecp-endpoint or /etc/edx-toolbox)");
    System.out.println(" QUEUES        :                    : Number of different options to specify the queues to scan: ");
    System.out.println("                 (<queue-name>[,<queue-name>]*)    : will scan these queues");
    System.out.println("                 <edx.yml>                         : path to edx.yml, will scan ALL queues from edx.yml");
    System.out.println("                 <ecp.properties>                  : path to ecp.properties, will scan a fixed list of ecp-queues");
    System.out.println(" FILTER        : (<messageType>|ALL)[,TTL-<sec>][,SEN-<senderCode>][,REC-<receiverCode>][,BAU-<businessApp>] - see below for expl.");
    System.out.println("                 <messageType>|ALL  : Mandatory filter - will match both on businessType and messageType. Keyword 'ALL' matches all msg");
    System.out.println("                 TTL-<sec>          : Optional filter to specify max <sec> TTL/age for the message - if omitted it matches all msg");
    System.out.println("                 SEN-<senderCode>   : Optional filter to specify <senderCode> for the message - if omitted it matches all msg");
    System.out.println("                 REC-<receiverCode> : Optional filter to specify <receiverCode> for the message - if omitted it matches all msg");
    System.out.println("                 BAU-<businessApp>  : Optional filter to specify <BA> for the message - if omitted it matches all msg");
    System.out.println(" KEYSTORE      : <keystore-file> <keystore-password> - these two parameters are only necessary if you specify 'amqps' as URL argument. ");
    System.out.println("                                                       If they are specified they will override keystore params found in PATH-config");

    System.out.println("\nExample 1: Consume all messages from standard ActiveMQ DLQ:");
    System.out.println("\tjava -jar ConsumerExample.jar ALL amqp://localhost:5672 ActiveMQ.DLQ ALL");

    System.out.println("\nExample 2: Consume older than 1h messages on edx.endpoint.reply (case: typically leftover msg that BA is not consuming):");
    System.out.println("\tjava -jar ConsumerExample.jar ALL amqp://localhost:5672 edx.endpoint.reply ALL,TTL-3600");

    System.out.println("\nExample 3: Consume older than 1h messages on ecp.endpoint.inbox and ecp.endpoint.send.event (case: msg stuck on-route from ECP to EDX):");
    System.out.println("\tjava -jar ConsumerExample.jar ALL amqp://localhost:5672 ecp.endpoint.inbox,ecp.endpoint.send.event ALL,TTL-3600");

    System.out.println("\nExample 4: Consume max 40 message of specific type from DLQ (use AMQPS) (case: consume 1d old AOL-msg from DLQ sent from ECO-2 endpoint):");
    System.out.println("\tjava ConsumerExample.jar 40 amqps://toolbox:password@localhost:5672 ActiveMQ.DLQ NBM-CIM-PTZ12-MTZ35,TTL-86400,SEN-50V-NBM-E2HT-ATP authKeystore.jks password");

    System.out.println("\nExample 5: Browse first 400 messages from standard ActiveMQ DLQ:");
    System.out.println("\tjava -jar ConsumerExample.jar NONE amqp://localhost:5672 ActiveMQ.DLQ ALL");

    System.out.println("\nExample 6: Browse first 400 messages from all queues on EDX:");
    System.out.println("\tjava -jar ConsumerExample.jar NONE /etc/edx-toolbox ALL ALL");

    System.out.println("\nExample 7: Browse first 400 10-min old messages from all queues found in edx.yml:");
    System.out.println("\tjava -jar ConsumerExample.jar NONE amqp://localhost:6672 /etc/edx-toolbox/edx.yml ALL,TTL-600");

    System.out.println("If the ConsumerExample fails, take a look at the jndi.properties file that it makes and examine the URL it uses");
  }

}