/*
 * Copyright (C) Red Hat, Inc.
 * http://www.redhat.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package no.statnett.ecp.burst;

import no.statnett.ecp.AMQPConnector;

import javax.jms.*;
import javax.naming.Context;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;

public class BurstProducer {

  private static String VERSION = "v0.9.0";

  public static void main(String args[]) throws IOException {

    if (args.length < 7 || args.length > 9) {
      usage();
      System.exit(1);
    }

    String brokerArg = args[0];
    String queue = args[1];
    String receiverCode = args[2];
    String messageType = args[3];
    String file = args[4];
    int repeated = Integer.parseInt(args[5]);
    int intervalMs = Integer.parseInt(args[6]);
    String keystoreFile = null; // specify file if using amqps in urlPathArg, example if connecting to ECP: "/var/lib/ecp-endpoint/authKeystore.jks"
    String keystorePW = null; // specify file if using amqps in urlPathArg, example: "password"
    if (args.length > 7) {
      keystoreFile = args[7];
      keystorePW = args[8];
    }


    // The message content/payload
    byte[] messageFromArg = Files.readAllBytes(Path.of(file));

    try {
      Map<String, String> config = AMQPConnector.findBrokerConfig(brokerArg, keystoreFile, keystorePW);
      Context context = AMQPConnector.getContext(true, config, queue);
      Connection connection = AMQPConnector.createSession(context);
      Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
      MessageProducer producer = session.createProducer((Destination) context.lookup(AMQPConnector.JNDI_QUEUE_REFERENCE));
      BytesMessage message = AMQPConnector.buildBytesMessage(session, messageFromArg, receiverCode, messageType);
      String sizeStr = messageFromArg.length < 1024 ? messageFromArg.length + "B" : messageFromArg.length < 1024 * 1024 ? messageFromArg.length / 1024 + "KB" : messageFromArg.length / (1024 * 1024) + "MB";
      String intervalStr = intervalMs == 0 ? "no interval" : "interval of " + intervalMs + "ms";
      long printMs = System.currentTimeMillis();
      System.out.println("Starting. Will send " + (repeated + 1) + " messages of size " + sizeStr + " with " + intervalStr + " going to " + receiverCode);
      for (int i = 0; i < repeated + 1; i++) {
        producer.send(message);
        if (intervalMs > 0)
          Thread.sleep(intervalMs);
        if (System.currentTimeMillis() - printMs > 10000) {
          System.out.println("Sent " + i + " message(s)");
          printMs = System.currentTimeMillis();
        }
      }
      producer.close();
      System.out.println("Completed. Sent " + (repeated + 1) + " message(s)");
    } catch (Throwable t) {
      System.err.println("Error sending message" + t);
    }
  }

  private static void usage() {
    System.out.println("ProducerExample " + VERSION + " is a demonstration on how to produce to ECP/EDX over AMQP(S). It can be used like this:\n\n");
    System.out.println("Usage  : java -jar ProducerExample.jar BROKER <queue-name> <receiver> <message-type> <file> <repeated> <interval-ms> [KEYSTORE]\n");
    System.out.println("\n\nThe arguments:");
    System.out.println(" BROKER        : <URL|PATH> - either specify a URL to a broker or a path for the ECP- or EDX-configuration");
    System.out.println("                 URL                : Specify url starting with either 'amqp' or 'amqps' (the latter requires extra args - see below");
    System.out.println("                 PATH               : Specify either the conf-path for ECP or EDX (ex: /etc/ecp-endpoint or /etc/edx-toolbox)");
    System.out.println(" <receiver>    : Specify the receiver code/address");
    System.out.println(" <message-type>: Specify the message-type/business-type");
    System.out.println(" <file>        : Specify filename to send");
    System.out.println(" <repeated>    : Specify number of times to repeat the producing of the same file to queue. \"0\" means file is sent once.");
    System.out.println(" <interval-ms> : Specify interval in milliseconds between repeated sending.");
    System.out.println(" KEYSTORE      : <keystore-file> <keystore-password> - these two parameters are only necessary if you specify 'amqps' as URL argument. ");
    System.out.println("                                                       If they are specified they will override keystore params found in PATH-config");

    System.out.println("\nExample 1: Send one TEST-message to 50V000000000115O via the standard ECP outbox using local AMQP:");
    System.out.println("\tjava -jar ProducerExample.jar amqp://localhost:5672 ecp.endpoint.outbox 50V000000000115O TEST small.txt 0 0");

    System.out.println("\nExample 2: Send 100 TEST-messages to 50V000000000115O via the standard ECP outbox using local AMQP, 1 ms delay between each message");
    System.out.println("\tjava -jar ProducerExample.jar amqp://localhost:5672 ecp.endpoint.outbox 50V000000000115O TEST small.txt 99 1");

    System.out.println("\nExample 3: Send one EXAMPLE-message to the FSSF-service via the standard EDX outbox using local AMQP:");
    System.out.println("\tjava -jar ProducerExample.jar /etc/edx-toolbox edx.endpoint.outbox SERVICE-FSSF EXAMPLE small.txt 0 0");

    System.out.println("\nExample 4: Send one TEST-message to 50V000000000115O via the standard ECP outbox using remote AMQPS:");
    System.out.println("\tjava -jar ProducerExample.jar amqps://endpoint:password@hostname.com:5672 ecp.endpoint.outbox 50V000000000115O TEST small.txt 0 0 keystores/authKeystore.jks password");

    System.out.println("\nExample 5: Send 20 TIMEOUT-messages to 50V000000000115O via the standard EDX outbox using remote AMQP, 0 ms delay between each message");
    System.out.println("\tjava -jar ProducerExample.jar amqp://hostname.com:6672 edx.endpoint.outbox 50V000000000115O TIMEOUT big.zip 19 0");

    System.out.println("If the ProducerExample fails, take a look at the jndi.properties file that it makes and examine the URL it uses");
  }

}
