package no.statnett.ecp.log;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.GZIPInputStream;

import static java.time.ZoneOffset.UTC;


public class ECPMsgDelay {

  private static final DateTimeFormatter tmsF = DateTimeFormatter.ofPattern("yyyyMMddHHmmssSSS");
  private static final DateTimeFormatter tmsHourF = DateTimeFormatter.ofPattern("yyyyMMddHH");
  private static final DateTimeFormatter dateF = DateTimeFormatter.ofPattern("yyyyMMdd");
  private static final DateTimeFormatter jsonF = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS'Z'");
  private static long oldestEcpTms = Long.MAX_VALUE;
  private static long newestEcpTms = 0;

  private static Map<String, Message> sndMsgMap = new HashMap<>();
  private static Map<String, Message> rcvMsgMap = new HashMap<>();

  private static Map<String, Long> mapRCplusGTMStoPayloadSize = new HashMap<>();

  private static boolean prevHour = false; // default
  private static boolean initialize = false; // default
  private static int delayLim = 10;
  private static int expireLim = 60;
  private static int noDaysSummary = 7;
  private static boolean snd = false;
  private static boolean rcv = false;
  private static File limitFile;
  private static File clockSkewFile;
  private static File oppositeFile;
  private static File summaryFile;


  public static void main(String[] args) {
    args = parseOptions(args);
    try {
      String ecpLog = null;
      if (args.length != 1) {
        if (summaryFile == null) {
          usage();
          System.exit(1);
        }
      } else {
        ecpLog = args[0];
      }

      if (summaryFile == null) { // default behavior
        processDirectory(ecpLog);

        Long oldestTms = oldestEcpTms;
        Long newestTms = newestEcpTms;

        // Parse oldestTms and newestTms into ZonedDateTime
        ZonedDateTime oldest = LocalDateTime.parse("" + oldestTms, tmsF).atZone(UTC);
        ZonedDateTime newest = LocalDateTime.parse("" + newestTms, tmsF).atZone(UTC);
        ZonedDateTime startTms = oldest.truncatedTo(ChronoUnit.MINUTES);
        ZonedDateTime endTms = newest.truncatedTo(ChronoUnit.MINUTES);
        print(startTms, endTms);
      } else { // if -x is specified
        // Read first line of summary file
        String firstLine = Files.lines(summaryFile.toPath()).findFirst().orElse(null);
        if (firstLine == null) {
          System.err.println("Summary file is empty");
          System.exit(1);
        } else if (firstLine.contains("endpoint")) {
          processOppositeSummary(summaryFile);
        } else {
          processStandardSummary(summaryFile);
        }
      }
    } catch (Throwable t) {
      System.err.println("Error occurred: " + t);
    }
  }

  private static String[] parseOptions(String[] initialArgs) {
    List<String> args = new ArrayList<>();
    boolean nextArgIsOutputFile = false;
    boolean nextArgIsClockSkewFile = false;
    boolean nextArgIsOppositeFile = false;
    boolean nextArgIsSummaryfile = false;
    for (String initialArg : initialArgs) {
      if (nextArgIsClockSkewFile) {
        clockSkewFile = initiateFile(initialArg);
        nextArgIsClockSkewFile = false;
      } else if (nextArgIsOutputFile) {
        limitFile = initiateFile(initialArg);
        nextArgIsOutputFile = false;
      } else if (nextArgIsOppositeFile) {
        oppositeFile = initiateFile(initialArg);
        nextArgIsOppositeFile = false;
      } else if (nextArgIsSummaryfile) {
        summaryFile = initiateFile(initialArg);
        nextArgIsSummaryfile = false;
      } else if (initialArg.startsWith("-h")) {
        prevHour = true;
      } else if (initialArg.startsWith("-i")) {
        initialize = true;
      } else if (initialArg.startsWith("-d")) {
        delayLim = Integer.parseInt(initialArg.substring(2));
      } else if (initialArg.startsWith("-e")) {
        expireLim = Integer.parseInt(initialArg.substring(2));
      } else if (initialArg.startsWith("-y")) {
        noDaysSummary = Integer.parseInt(initialArg.substring(2));
      } else if (initialArg.startsWith("-c")) {
        nextArgIsClockSkewFile = true;
      } else if (initialArg.startsWith("-o")) {
        nextArgIsOutputFile = true;
      } else if (initialArg.startsWith("-x")) {
        nextArgIsSummaryfile = true;
      } else if (initialArg.startsWith("-p")) {
        nextArgIsOppositeFile = true;
      } else if (initialArg.startsWith("-r")) {
        rcv = true;
      } else if (initialArg.startsWith("-s")) {
        snd = true;
      } else {
        args.add(initialArg); // ecp-log-directory kommer til slutt ut her, men ikke når vi benytter -x
      }
    }
    if (rcv && snd || !rcv && !snd) { // default is snd = true, rcv= f
      snd = true;
      rcv = false;
    }
    if (expireLim <= delayLim) {
      expireLim = -1; // Det finnes ingen meldinger som regnes som "expired"
    }
    return args.toArray(new String[]{});
  }

  private static File initiateFile(String initialArgs) {
    File file = new File(initialArgs);
    try {
      if (file.exists()) {
        Files.writeString(file.toPath(), "", StandardOpenOption.APPEND);
      } else {
        Files.writeString(file.toPath(), "");
      }
      return file;
    } catch (IOException e) {
      System.err.println("Cannot write " + initialArgs + " (exception: " + e.getMessage() + ")");
      System.exit(1);
      return null; // System exists before this anyway
    }
  }


  private static void usage() {
    System.out.println("ECPMsgDelay v1.5.5");
    System.out.println("Usage  : java -jar ECPMsgDelay.jar [OPTION] <ecp-log-directory>\n");
    System.out.println(" OPTION:");
    System.out.println("     -s             : process only 'sent messages' - this is default behaviour, these timestamps are most trustworthy");
    System.out.println("     -r             : process only 'received messages', -s  will override -r");
    System.out.println("     -h             : all calculations is based on previous hour. Ideal for hourly cron-job");
    System.out.println("     -i             : all calculations is up-until previous hour. Ideal for first run (initialization)");
    System.out.println("     -d<sec>        : The delay-limit columns will show the number based on this limit in seconds. Default is 10");
    System.out.println("     -e<sec>        : The expire-limit columns will show the number based on this limit in seconds. Default is 60. Disabled if lower or equals to -d");
    System.out.println("     -o <filename>  : The entries that exceed the delay limit will be appended to this file");
    System.out.println("     -c <filename>  : Received entries that has been received before sent, indicating clock skew (only with -r)");
    System.out.println("     -p <filename>  : Append summary pr opposite endpoint pr hour to this file");
    System.out.println("     -x <filename>  : Produces monthly summary of the file specified. The file must contain the std-out or the opposite file (-p) from this program when run without -x.");
    System.out.println("     -y<nodays>     : Only valid with -x. Specifies the number of days to include in the summary. Default is last 7 days.");
    System.out.println("\nExample 1: Summary of all sent messages, delay/expire-limit is 10/60 sec:");
    System.out.println("\tjava -jar ECPMsgDelay.jar /var/log/ecp-endpoint");
    System.out.println("Example 2: Summary of last hour received messages, delay/expire-limit is 10/60 sec:");
    System.out.println("\tjava -jar ECPMsgDelay.jar -r -h /var/log/ecp-endpoint");
    System.out.println("Example 3: Summary of last hour received messages + 3 more files for different purposes, delay/expire-limit is 20/60 sec:");
    System.out.println("\tjava -jar ECPMsgDelay.jar -r -h -d20 -c neg.json -o lim.json -p opp.json /var/log/ecp-endpoint");
    System.out.println("Example 4: Summary of last hour sent messages + 2 files for different purposes, delay/expire-limit is 10/60 sec:");
    System.out.println("\tjava -jar ECPMsgDelay.jar -h -o lim.json -p opp.json /var/log/ecp-endpoint");
    System.out.println("Example 5: Initial summary, run before you start running hourly cron (-h):");
    System.out.println("\tjava -jar ECPMsgDelay.jar -i -o lim.json -p opp.json /var/log/ecp-endpoint");
    System.out.println("Example 6: Summary of summary, gives daily + monthly summary of the file specified (either stdout or opposite file):");
    System.out.println("\tjava -jar ECPMsgDelay.jar -x stdout.json -y7");
    System.out.println("\nThe delay analyzer is compatible with ECP 4.9-4.11 - it depends upon certain log events. Other");
    System.out.println("versions of ECP might not give any useful results. It calculates the delays of the message");
    System.out.println("transmissions in both directons, SEND and RECEIVE:");
    System.out.println("  * SEND delay is recorded when ACK is received and then compared with SEND tms");
    System.out.println("  * RECEIVE delay is recorded when STD message is recevied and compared with 'Generated' tms in the msg");
    System.out.println("\nThe main output is a series of json-objects, one pr hour with a summary of the delays/sizes.");
    System.out.println("The output is meant to be easy to read for humans and easy to parse for Splunk, etc. The OPTIONS allow you ");
    System.out.println("to investigate the delays in more detail.");
    System.out.println("\nSince generated tms inside the message (RECEIVE case) can come from an endpoint with a clock skewed");
    System.out.println("compared to this endpoint, we might get negative RECIEVE dealys. Those are not counted, but can be");
    System.out.println("listed using the -c option. The -o option will list all messages that exceed the limits in seconds.");
    System.out.println("The -p option will give a summary of the traffic for each opposite endpoint pr hour. Most data in the output");
    System.out.println("is self-explanatory, but the 'troubleIndex' is not. The index is calculated like this:");
    System.out.println("  troubleIndex = overrepresentation * limitFreqOfEndpoint * limitCount * limitAvgIndex / 100");
    System.out.println("     overrepresentation is 100 if limits are not overrepresented nor underrepresented");
    System.out.println("     limitFreqOfEndpoint is 100 if all traffic to the endpoint exceed the limit");
    System.out.println("     limitCount is simply the number of messages that exceed the limit");
    System.out.println("     limitAvgIndex is the index (1-100), 1 when limitAvg == limit up to max 100 when limitAvg >= expireLimit (60s default)");
    System.out.println("The higher the index, the worse the problem is.");
  }

  private static void processDirectory(String dir) throws IOException {
    File fileDir = new File(dir);
    if (fileDir.isDirectory()) {
      for (Path p : Files.list(Path.of(dir)).filter(p -> p.getFileName().toString().matches("ecp.(log|.*gz)|ecp-audit(.*log|.*gz)")).sorted((o1, o2) -> (int) (o1.toFile().lastModified() - o2.toFile().lastModified())).collect(Collectors.toList())) {
        try {
          if (p.toString().endsWith("gz")) {
            processGZIP(p);
          } else {
            processPlainText(p.toString());
          }
        } catch (IOException e) {
          e.printStackTrace();
        }
      }
    } else {
      System.err.println("Input argument " + fileDir + " is not a directory - exiting");
      System.exit(1);
    }
  }

  private static void processGZIP(Path p) throws IOException {
    InputStream fileStream = new FileInputStream(p.toFile());
    InputStream gzipStream = new GZIPInputStream(fileStream);
    Reader decoder = new InputStreamReader(gzipStream, StandardCharsets.UTF_8);
    new BufferedReader(decoder).lines().forEach(s -> processLogLine(s));
  }

  private static void processPlainText(String log) {
    try (Stream<String> stream = Files.lines(Paths.get(log), StandardCharsets.UTF_8)) {
      stream.forEach(s -> {
        processLogLine(s);
      });
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  // For å matche meldinger som sendes med ACK'er som logges (meldinger vi sender)

  // 1. Linje som inneholder "MSG-11" og "STANDARD_MESSAGE" markerer starten på å sende melding
  //    Søk messageID=(.*), => ID for meldingen
  // Eksempel: 2023-02-27 12:48:12.996  INFO 126224 --- [Camel (camel-3) thread #67 - JmsConsumer[ecp.endpoint.internal.outbox]] EcpAuditLog           : ComponentCode=50V-SN-TEST--ATK, ComponentType=ECP Endpoint, RecordID=3528b69f-4e23-45ab-a620-2595c72ea160, ID=MSG-11,  Category=MSG,  Text=Message [messageID=c372ed5a-0f96-4b3a-b9ec-794fba01d826,internalType=STANDARD_MESSAGE,senderCode=50V-SN-TEST--ATK,receiverCode=50V-SN-DK----ATT,businessType=MORTEN-TEST,baMessageID=<null>,senderApplication=MORTEN,generated=2023-02-27T13:48:12.905+01:00[GMT+01:00]] has been uploaded to 50V000000000112U., Direction=OUT, User=SYSTEM, Severity=0

  // 2. Linje som inneholder "MSG-09" eller "MSG-10" inneholder ACK
  //    Søk messageID=(.*), => ID for meldingen
  // Eksempel: 2023-02-27 12:48:13.338  INFO 126224 --- [Camel (camel-3) thread #66 - JmsConsumer[ecp.endpoint.download]] EcpAuditLog                  : ComponentCode=50V-SN-TEST--ATK, ComponentType=ECP Endpoint, RecordID=414a47d7-2569-46af-9b6b-e2c9f3742add, ID=MSG-09,  Category=MSG,  Text=Message [messageID=c372ed5a-0f96-4b3a-b9ec-794fba01d826,internalType=STANDARD_MESSAGE,senderCode=50V-SN-TEST--ATK,receiverCode=50V-SN-DK----ATT,businessType=MORTEN-TEST,baMessageID=<null>,senderApplication=MORTEN,generated=<null>] has been delivered., Direction=OUT, User=SYSTEM, Severity=0

  // For å finne meldingers om vi mottar:
  // 3. Linje som inneholder "MSG-08" og "STANDARD_MESSAGE" markerer starten på å motta melding
  // Eksempel: 2023-03-07 08:34:35.670  INFO 126224 --- [Camel (camel-3) thread #81 - JmsConsumer[ecp.endpoint.receive.50V-SN-TEST--ATK]] EcpAuditLog                              : ComponentCode=50V-SN-TEST--ATK, ComponentType=ECP Endpoint, RecordID=bf4e2277-d0db-41ef-902b-bac262380220, ID=MSG-08, Category=MSG, Text=Message [messageID=b2317a15-e594-46a5-ab7d-bdaaa8416ef7,internalType=STANDARD_MESSAGE,senderCode=50V000000000115O,receiverCode=50V-SN-TEST--ATK,businessType=TEST,baMessageID=<null>,senderApplication=MORTEN,generated=2023-03-07T09:34:35.467+01:00[GMT+01:00]] has been received from 50V000000000112U., Direction=IN, User=SYSTEM, Severity=0

  // Noe senere...
  // Det måtte selvsagt bli litt mer komplisert da vi ønsket å hente ut payloadSize knyttet til hver meldingstype. Riktignok finner vi denne på MSG-01 (send) og MSG-02 (receive), men for MSG-01 så er ikke messageID fylt ut ennå!! Derfor må vi gjøre en rekke krumspring for å koble payloadSize til korrekt messageID. Dette gjør
  // vi gjennom å hente ut generatedTms + receiverCode i MSG-11 og så finne MSG-01 som har samme generatedTms + receiverCode. Da har vi messageID-match og kan hente ut payloadSize fra MSG-01 og koble den til rett messageId. For MSG-02 er dette rett fram, fordi der har vi messageID allerede fylt ut. Jeg gidder ikke å vise
  // loglinjene her - men de fungerer for 4.10 og 4.11 av ECP.

  // s is the line we'll investigate/parse
  private static void processLogLine(String s) {
    try {
      if (s != null && s.length() > 20 && (s.startsWith("2023") || s.startsWith("2024"))) {
        long tms = Long.parseLong(s.substring(0, 23).replaceAll("[^\\d]", ""));

        // Man finner først MSG-01 (men den inneholder ikke msgId), deretter MSG-11 (m/ msgId) (ECP forsøker å sende en melding), deretter MSG-09/10 (ACK som kommer tilbake m/msgId)
        if (snd) {
          if (s.contains("MSG-01") && s.contains("STANDARD_MESSAGE") && s.contains("payloadSize=")) { // 1.
            long payloadSize = findPayloadSizeInLogLine(s);
            long generatedTms = findGeneratedTmsInLogLine(s);
            String receiverCode = findReceiverCodeInLogLine(s);
            String key = "" + generatedTms + receiverCode;
            mapRCplusGTMStoPayloadSize.put(key, payloadSize); // nøkkelen "receiverCode + generatedTms" brukes for å koble payloadSize til riktig messageID i MSG-11-seksjonen
          } else if (s.contains("MSG-11") && s.contains("STANDARD_MESSAGE")) { // 1.
            String messageId = findMessageIDInLogLine(s);
            Message message = new Message(messageId, true);
            message.setTmpTms(LocalDateTime.parse("" + tms, tmsF).toInstant(UTC).toEpochMilli());
            message.setBroker(findUploadBrokerInLogLine(s));
            long generatedTms = findGeneratedTmsInLogLine(s);
            String receiverCode = findReceiverCodeInLogLine(s);
            String key = "" + generatedTms + receiverCode;
            Long payloadSize = mapRCplusGTMStoPayloadSize.get(key); // finner nøkkelen "receiverCode + generatedTms" og bruker den til å slå opp payloadSize som ble laget over (i MSG-01-seksjonen) (vi er da noenlunde trygge på at dette gjelder samme msgId)
            if (payloadSize != null) {
              message.setBytes(payloadSize);
            }
            sndMsgMap.put(messageId, message);
          } else if (s.contains("MSG-09") || s.contains("MSG-10")) { // 2.
            String messageId = findMessageIDInLogLine(s);
            Message message = sndMsgMap.get(messageId);
            if (message != null) {
              message.setTms(LocalDateTime.parse("" + tms, tmsF).toInstant(UTC).toEpochMilli());
              message.setMs(message.getTms() - message.getTmpTms());
              message.setTmsHour(Long.parseLong(ZonedDateTime.ofInstant(Instant.ofEpochMilli(message.getTms()), UTC).format(tmsHourF)));
              message.setReceiver(findReceiverCodeInLogLine(s));
              message.setSender(findSenderCodeInLogLine(s));
            }
          }
        } else { // rcv == true
          // Man finner først MSG-02 (her finner vi payload + messageId), deretter kompletteres mottaket av en melding med MSG-08 - denne er litt mer straight-forward enn når man sender en melding ut, fordi vi har meldingsId+payloadSize sammen
          if (s.contains("MSG-08") && s.contains("STANDARD_MESSAGE")) {
            String messageId = findMessageIDInLogLine(s);
            Message message = rcvMsgMap.get(messageId);
            if (message == null) {
              message = new Message(messageId, false);
              rcvMsgMap.put(messageId, message);
            }
            message.setBroker(findDownloadBrokerInLogLine(s));
            rcvMsgMap.put(messageId, message);
          } else if (s.contains("MSG-02") && s.contains("STANDARD_MESSAGE") && s.contains("payloadSize=")) { // 3.
            long payloadSize = findPayloadSizeInLogLine(s);
            String messageId = findMessageIDInLogLine(s);
            Message message = rcvMsgMap.get(messageId);
            if (message == null) {
              message = new Message(messageId, false);
              rcvMsgMap.put(messageId, message);
            }
            message.setBytes(payloadSize);
            long rcvStartEpochMs = findGeneratedTmsInLogLine(s);
            message.setTmpTms(rcvStartEpochMs);
            long rcvEndEcpochMs = LocalDateTime.parse("" + tms, tmsF).toInstant(UTC).toEpochMilli();
            message.setTms(rcvEndEcpochMs);
            message.setMs(message.getTms() - message.getTmpTms());
            message.setTmsHour(Long.parseLong(ZonedDateTime.ofInstant(Instant.ofEpochMilli(message.getTms()), UTC).format(tmsHourF)));
            message.setReceiver(findReceiverCodeInLogLine(s));
            message.setSender(findSenderCodeInLogLine(s));
          }
        }

        if (tms < oldestEcpTms) {
          oldestEcpTms = tms;
        }
        if (tms > newestEcpTms)
          newestEcpTms = tms;
      }
    } catch (NumberFormatException |
      DateTimeParseException ex) {
      System.err.println("Error occurred: " + ex + " with input text: " + s);
    }
  }

  private static String findMessageIDInLogLine(String s) {
    int idPosStart = s.indexOf("messageID=") + "messageId=".length(); //
    int idPosSlutt = s.indexOf(",", idPosStart);
    return s.substring(idPosStart, idPosSlutt);
  }

  private static String findUploadBrokerInLogLine(String s) {
    int idPosStart = s.indexOf("uploaded to ") + "uploaded to ".length(); //
    int idPosSlutt = s.indexOf(".,", idPosStart);
    return s.substring(idPosStart, idPosSlutt);
  }

  // ...has been received from 50V000000000112U.,...
  private static String findDownloadBrokerInLogLine(String s) {
    int idPosStart = s.indexOf("has been received from ") + "has been received from ".length(); //
    int idPosSlutt = s.indexOf(".,", idPosStart);
    return s.substring(idPosStart, idPosSlutt);
  }

  private static String findReceiverCodeInLogLine(String s) {
    int idPosStart = s.indexOf("receiverCode=") + "receiverCode=".length(); //
    int idPosSlutt = s.indexOf(",", idPosStart);
    return s.substring(idPosStart, idPosSlutt);
  }

  private static String findSenderCodeInLogLine(String s) {
    int idPosStart = s.indexOf("senderCode=") + "senderCode=".length(); //
    int idPosSlutt = s.indexOf(",", idPosStart);
    return s.substring(idPosStart, idPosSlutt);
  }

  // All these handles by ZonedDateTime
  // ...generated=2023-03-07T09:34:35.467+01:00[GMT+01:00]]....
  // ...generated=2023-03-07T09:34:35+01:00[GMT+01:00]]....
  // ...generated=2023-03-02T22:51:19.917Z[GMT]
  // ...generated=2023-03-03T04:50:54Z[GMT]
  private static Long findGeneratedTmsInLogLine(String s) {
    int idPosStart = s.indexOf("generated=") + "generated=".length(); //
    int idPosSlutt = s.indexOf("]]", idPosStart) + 1; // inkluderer "]" i tms
    String tmsStr = s.substring(idPosStart, idPosSlutt);
    return ZonedDateTime.parse(tmsStr).toInstant().toEpochMilli();
  }

  private static long findPayloadSizeInLogLine(String s) {
    int idPosStart = s.indexOf("payloadSize=") + "payloadSize=".length();
    int idPosSlutt = s.indexOf(".,", idPosStart);
    return Long.parseLong(s.substring(idPosStart, idPosSlutt));

  }

  // Vi har hentet inn alle data vi trenger, men vi skriver bare ut den del som er relevant, gitt de argumenter vi har fått

  private static void print(ZonedDateTime startTms, ZonedDateTime endTms) throws IOException {
    // Make the timestamp "2023-03-03 14:04" into ZonedDateTime UTC
//    ZonedDateTime matchTms = ZonedDateTime.of(2023, 3, 3, 14, 1, 0, 0, UTC);
    ZonedDateTime matchTms = Instant.now().atZone(UTC);
    int matchDate = 0;
    if (prevHour || initialize) {
      matchTms = matchTms.minusHours(1).truncatedTo(ChronoUnit.HOURS);
      matchDate = Integer.parseInt(matchTms.format(dateF));
    }

    for (ZonedDateTime tms = startTms; tms.compareTo(endTms) < 0; tms = tms.truncatedTo(ChronoUnit.HOURS).plusHours(1)) {
      if (prevHour) { // -h is set, will override -i
        boolean tmsPrevHour = Integer.parseInt(tms.format(dateF)) == matchDate && tms.getHour() == matchTms.getHour();
        if (!tmsPrevHour) {
          continue;// skip if not the previousHour AND previousHour-only is requested - useful for crontab that runs every hour and print one single hour
        }
      } else { // -h is not set, but -i could or could not be set
        if (initialize) {
          boolean afterPrevHour = tms.compareTo(matchTms) >= 0;
          if (afterPrevHour) {
            continue; // skip if after previousHour - useful for initial run (and then run hourly with -h later)
          }
        }
      }
      long hourKey = Long.parseLong(tms.format(tmsHourF));
      Map<String, Message> msgMap = snd ? sndMsgMap : rcvMsgMap;

      // Main map with all messages
      Map<String, Message> hourMap = msgMap.values().stream().filter(m -> m.getTmsHour() == hourKey && m.getMs() >= 0).collect(Collectors.toMap(Message::getMessageId, m -> m));
      // Main map with all messages that are above the delayLimit
      Map<String, Message> delayLimitMap = hourMap.values().stream().filter(m -> m.getMs() >= delayLim * 1000L).collect(Collectors.toMap(Message::getMessageId, m -> m));
      Map<String, Message> expireLimitMap = expireLim == -1 ? new HashMap<>() : hourMap.values().stream().filter(m -> m.getMs() >= expireLim * 1000L).collect(Collectors.toMap(Message::getMessageId, m -> m));

      // Default output to stdout
      System.out.print(logHOURSummary(tms, hourMap, delayLimitMap, expireLimitMap, snd));

      if (limitFile != null) { // Optional output to file - delayLimits
        logDelayLimits(delayLimitMap);
      }
      if (oppositeFile != null) { // Optional output to file - opposite
        logOpposite(tms, hourMap, delayLimitMap);
      }

      if (clockSkewFile != null && rcv) { // Optional output to file - clock skew
        Map<String, Message> rcvNegMap = hourMap.values().stream().filter(m -> m.getMs() < 0).collect(Collectors.toMap(Message::getMessageId, m -> m));
        logClockSkew(rcvNegMap);
      }
    }
    return;
  }

  private static StringBuffer logHOURSummary(ZonedDateTime tms, Map<String, Message> hourMap, Map<String, Message> delayLimitMap, Map<String, Message> expireLimitMap, boolean snd) {
    long cntTot = hourMap.size();
    long cntExp = expireLimitMap.size();
    long cntDel = delayLimitMap.size() - cntExp;

    long avgMs = (long) hourMap.values().stream().mapToLong(Message::getMs).average().orElse(0);
    long msExp = (long) expireLimitMap.values().stream().mapToLong(Message::getMs).sum();
    long msDel = (long) delayLimitMap.values().stream().mapToLong(Message::getMs).sum();
    long avgMsDel = cntDel == 0 ? 0 : (msDel - msExp) / cntDel;
    long avgMsExp = cntExp == 0 ? 0 : msExp / cntExp;

    long avgKB = (long) hourMap.values().stream().mapToLong(Message::getKB).average().orElse(0);
    long kbExp = (long) expireLimitMap.values().stream().mapToLong(Message::getKB).sum();
    long kbDel = (long) delayLimitMap.values().stream().mapToLong(Message::getKB).sum();
    long avgKBDel = cntDel == 0 ? 0 : (kbDel - kbExp) / cntDel;
    long avgKBExp = cntExp == 0 ? 0 : kbExp / cntExp;

    long maxMs = hourMap.values().stream().mapToLong(Message::getMs).max().orElse(0);
    long maxKB = hourMap.values().stream().mapToLong(Message::getKB).max().orElse(0);

    StringBuffer sb = new StringBuffer();
    sb.append("{ \"timestamp\": \"").append(tms.format(jsonF)).append("\", ").
      append("\"direction\": \"").append(snd ? "snd" : "rcv").append("\", ").
      append("\"cnt\": ").append(String.format("%4d", cntTot)).append(", ").
      append("\"cntDel\": ").append(String.format("%4d", cntDel)).append(", ").
      append("\"cntExp\": ").append(String.format("%4d", cntExp)).append(", ").

      append("\"avgMs\": ").append(String.format("%9d", avgMs)).append(", ").
      append("\"avgMsDel\": ").append(String.format("%5d", avgMsDel)).append(", ").
      append("\"avgMsExp\": ").append(String.format("%9d", avgMsExp)).append(", ").

      append("\"avgKB\": ").append(String.format("%5d", avgKB)).append(", ").
      append("\"avgKBDel\": ").append(String.format("%5d", avgKBDel)).append(", ").
      append("\"avgKBExp\": ").append(String.format("%5d", avgKBExp)).append(", ").

      append("\"maxMs\": ").append(String.format("%10d", maxMs)).append(", ").
      append("\"maxKB\": ").append(String.format("%5d", maxKB)).
      append(" }\n");
    return sb;
  }

  private static void logDelayLimits(Map<String, Message> delayLimitMap) throws IOException {
    StringBuffer sb = new StringBuffer();
    for (Message message : delayLimitMap.values().stream().sorted(Comparator.comparingLong(Message::getTms)).collect(Collectors.toList())) {
      // convert tms to a string on the form of "yyyy-MM-dd HH:mm:ss"
      sb.append("{ \"timestamp\": \"").append(ZonedDateTime.ofInstant(Instant.ofEpochMilli(message.getTms()), UTC).format(jsonF)).append("\", ").
        append("\"direction\": \"").append(message.getDirection()).append("\", ").
        append("\"msgId\": \"").append(message.getMessageId()).append("\", ").
        append("\"ms\": ").append(String.format("%6d", message.getMs())).append(", ").
        append("\"KB\": ").append(String.format("%5d", message.getKB())).append(", ").
        append("\"KBs\": ").append(String.format("%5d", (1000L * message.getKB()) / message.getMs())).append(", ").
        append("\"measureEP\": \"").append(message.getMeasureEP()).append("\", ").
        append("\"broker\": \"").append(message.getBroker()).append("\", ").
        append("\"oppositeEP\": \"").append(message.getOppositeEP()).append("\"").
        append(" }\n");
    }
    Files.writeString(limitFile.toPath(), sb.toString(), StandardOpenOption.APPEND);
  }

  private static void logClockSkew(Map<String, Message> negMap) throws IOException {
    if (negMap.size() > 0) {
      StringBuffer sb = new StringBuffer();
      for (Message message : negMap.values().stream().sorted(Comparator.comparingLong(Message::getTms)).collect(Collectors.toList())) {
        // convert tms to a string on the form of "yyyy-MM-dd HH:mm:ss"
        sb.append("{ \"timestamp\": \"").append(ZonedDateTime.ofInstant(Instant.ofEpochMilli(message.getTms()), UTC).format(jsonF)).append("\", ").
          append("\"direction\": \"").append(message.getDirection()).append("\", ").
          append("\"msgId\": \"").append(message.getTms()).append("\", ").
          append("\"ms\": ").append(String.format("%6d", message.getMs())).append(", ").
          append("\"measureEP\": \"").append(message.getMeasureEP()).append("\", ").
          append("\"oppositeEP\": \"").append(message.getOppositeEP()).append("\"").
          append(" }\n");
      }
      Files.writeString(clockSkewFile.toPath(), sb.toString(), StandardOpenOption.APPEND);
    }
  }

  private static void logOpposite(ZonedDateTime tms, Map<String, Message> hourMap, Map<String, Message> delayLimitMap) throws IOException {

    // Prepare summary maps (counts, delay, sizes) for all messages in an hour
    Map<String, Integer> hourCountMap = hourMap.values().stream().collect(Collectors.groupingBy(Message::getOppositeEP, Collectors.summingInt(m -> 1)));
    int totalCount = hourCountMap.values().stream().mapToInt(Integer::intValue).sum();
    Map<String, Long> hourMsMap = hourMap.values().stream().collect(Collectors.groupingBy(Message::getOppositeEP, Collectors.summingLong(Message::getMs)));
    Map<String, Long> hourKBMap = hourMap.values().stream().collect(Collectors.groupingBy(Message::getOppositeEP, Collectors.summingLong(Message::getKB)));

    // Prepare summary maps (counts, delay, sizes) for all messages in an hour that are above the delayLimit
    Map<String, Integer> delayLimitCountMap = delayLimitMap.values().stream().collect(Collectors.groupingBy(Message::getOppositeEP, Collectors.summingInt(m -> 1)));
    int totalDelayLimitCount = delayLimitCountMap.values().stream().mapToInt(Integer::intValue).sum();
    Map<String, Long> delayLimitMsMap = delayLimitMap.values().stream().collect(Collectors.groupingBy(Message::getOppositeEP, Collectors.summingLong(Message::getMs)));
    Map<String, Long> delayLimitKBMap = delayLimitMap.values().stream().collect(Collectors.groupingBy(Message::getOppositeEP, Collectors.summingLong(Message::getKB)));


    StringBuffer sb = new StringBuffer();
    for (String oppositeEP : hourCountMap.keySet()) {
      long count = hourCountMap.get(oppositeEP); // count always greater than 0
      long avgMs = hourMsMap.get(oppositeEP) / count;
      long avgKB = hourKBMap.get(oppositeEP) / count;
      long avgKBs = avgMs == 0 ? 0 : (1000L * avgKB) / avgMs;
      long freqAll = (100L * count) / totalCount; // totalCount always greater than 0 - range 0-100

      long countLim = delayLimitCountMap.getOrDefault(oppositeEP, 0); // how many delayLimits to this endpoint?
      long avgMsLim = countLim == 0 ? 0 : delayLimitMsMap.getOrDefault(oppositeEP, 0L) / countLim;
      long avgKBLim = countLim == 0 ? 0 : delayLimitKBMap.getOrDefault(oppositeEP, 0L) / countLim;
      long avgKBsLim = avgMsLim == 0 ? 0 : (1000L * avgKBLim) / avgMsLim;

      long freqDelayLimitAll = totalDelayLimitCount == 0 ? 0 : (100L * countLim) / totalDelayLimitCount;
      long delayLimitOverrepresentation = freqAll == 0 ? 0 : (100L * freqDelayLimitAll) / freqAll; // are there more delayLimits than to be expected based on the overall traffic to this endpoint? (100% = as many delayLimits as expected)
      long freqDelayLimitSelf = (100L * countLim) / count; // how much delayLimits compared to all messages to this endpoint (100% = as many delayLimits as messages)
      int delayLimitMs = delayLim * 1000;
      long avgLimLimDiff = avgMsLim - delayLimitMs; // difference between the average of those messages exceeding the delayLimit and the delayLimit itself
      long maxExpire = Math.max(expireLim, delayLim) * 1000L;
      long delayLimitAvgIndex = avgLimLimDiff <= 0 ? 0 : (avgLimLimDiff >= maxExpire ? 100 : 1 + ((99L * avgLimLimDiff) / maxExpire)); // 0 = no delay, 1 = average delay equals delayLimit, 100 = 10 minutes delay or more on average (for those exceeding the delayLimit)
      long troubleIndex = (delayLimitOverrepresentation * freqDelayLimitSelf * countLim * delayLimitAvgIndex) / 100L;

      sb.append("{ \"timestamp\": \"").append(tms.format(jsonF)).append("\", ").
        append("\"endpoint\": \"").append(oppositeEP).append("\", ").
        append("\"direction\": \"").append(snd ? "snd" : "rcv").append("\", ").
        append("\"cnt\": ").append(String.format("%4d", count)).append(", ").
        append("\"avgMs\": ").append(String.format("%6d", avgMs)).append(", ").
        append("\"avgKB\": ").append(String.format("%5d", avgKB)).append(", ").
        append("\"avgKBs\": ").append(String.format("%5d", avgKBs)).append(", ").
        append("\"freq\": ").append(String.format("%3d", freqAll)).append(", ").

        append("\"cntLim\": ").append(String.format("%4d", countLim)).append(", ").
        append("\"avgMsLim\": ").append(String.format("%7d", avgMsLim)).append(", ").
        append("\"avgKBLim\": ").append(String.format("%5d", avgKBLim)).append(", ").
        append("\"avgKBsLim\": ").append(String.format("%5d", avgKBsLim)).append(", ").
        append("\"freqLim\": ").append(String.format("%3d", freqDelayLimitAll)).append(", ").
        append("\"freqLimSelf\": ").append(String.format("%3d", freqDelayLimitSelf)).append(", ").

        append("\"troubleIndex\": ").append(String.format("%9d", troubleIndex)).
        append(" }\n");
    }
    Files.writeString(oppositeFile.toPath(), sb.toString(), StandardOpenOption.APPEND);
  }

  private static void processStandardSummary(File summaryFile) {
    Locale.setDefault(Locale.US); // Sørger for . som desimalskilletegn

    TreeMap<String, StandardSummary> monthSummaryMap = new TreeMap<>(Comparator.naturalOrder());
    TreeMap<String, StandardSummary> dateSummaryMap = new TreeMap<>(Comparator.naturalOrder());
    try (Stream<String> stream = Files.lines(Paths.get(summaryFile.getPath()), StandardCharsets.UTF_8)) {
      stream.forEach(s -> {
        processSummaryLine(s, monthSummaryMap, dateSummaryMap);
      });

      // Skriver ut alle måneder
      for (String month : monthSummaryMap.keySet()) {
        System.out.print(monthSummaryMap.get(month).print());
      }

      // Litt logikk for å skrive ut de 7 siste dagene i den siste måneden
      List<StandardSummary> dateSummaryList = new ArrayList<>(dateSummaryMap.values());
      // Sorter listen basert på Summary.time-feltet
      dateSummaryList.sort(Comparator.comparing(StandardSummary::getTime));
      while (dateSummaryList.size() > noDaysSummary) {
        // Fjern første element i listen
        dateSummaryList.remove(0);
      }
      for (StandardSummary dateSummary : dateSummaryList) {
        System.out.print(dateSummary.print());
      }
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private static void processSummaryLine(String s, Map<String, StandardSummary> monthSummaryMap, Map<String, StandardSummary> dateSummaryMap) {
    if (s.contains("avgKBExp")) { // Inneholder nyeste versjon (1.0.0+) av ECPMsgDelay-output
      // { "timestamp": "2023-06-22 23:00:00.000Z", "direction": "snd", "cnt": 1527, "cntDel":    0, "cntExp":    0, "avgMs":    762, "avgMsDel":      0, "avgMsExp":      0, "avgKB":     6, "avgKBDel":     0, "avgKBExp":     0, "maxMs":    4625, "maxKB":    10 }
      int timestampStartPos = s.indexOf("timestamp") + 13;
      String month = s.substring(timestampStartPos, timestampStartPos + 7); // yyyy-MM
      String date = s.substring(timestampStartPos, timestampStartPos + 10); // yyyy-MM-dd
      int directionStartPos = s.indexOf("direction") + 13;
      String direction = s.substring(directionStartPos, directionStartPos + 3); // snd/rcv
      String line = s.substring(s.indexOf("cnt")); // resten av linjen sendes inn i Summary.add()
      monthSummaryMap.merge(month, new StandardSummary(month, direction, line), (oldSummary, newSummary) -> oldSummary.add(line));
      dateSummaryMap.merge(date, new StandardSummary(date, direction, line), (oldSummary, newSummary) -> oldSummary.add(line));
    }
  }

  private static void processOppositeSummary(File summaryFile) {
    Locale.setDefault(Locale.US); // Sørger for . som desimalskilletegn

    TreeMap<String, OppositeSummary> monthSummaryMap = new TreeMap<>(Comparator.naturalOrder());
    TreeMap<String, OppositeSummary> dateSummaryMap = new TreeMap<>(Comparator.naturalOrder());
    try (Stream<String> stream = Files.lines(Paths.get(summaryFile.getPath()), StandardCharsets.UTF_8)) {
      stream.forEach(s -> {
        processOppositeSummaryLine(s, monthSummaryMap, dateSummaryMap);
      });

      // Skriver ut alle måneder
      for (String month : monthSummaryMap.keySet()) {
        System.out.print(monthSummaryMap.get(month).print());
      }

      // Litt logikk for å skrive ut de 7 siste dagene i den siste måneden
      List<OppositeSummary> dateSummaryList = new ArrayList<>(dateSummaryMap.values());
      // Sorter listen basert på Summary.time-feltet
      dateSummaryList.sort(Comparator.comparing(OppositeSummary::getTime));
      while (dateSummaryList.size() > noDaysSummary) {
        // Fjern første element i listen
        dateSummaryList.remove(0);
      }
      for (OppositeSummary dateSummary : dateSummaryList) {
        System.out.print(dateSummary.print());
      }
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private static void processOppositeSummaryLine(String s, Map<String, OppositeSummary> monthSummaryMap, Map<String, OppositeSummary> dateSummaryMap) {
    if (s.contains("direction")) { // Inneholder nyeste versjon (1.5.1+) av ECPMsgDelay-output
      // { "timestamp": "2023-06-09 12:00:00.000Z", "direction": "snd", "endpoint": "50V000000000113S", "cnt":    1, "avgMs":    609, "avgKB":     0, "avgKBs":     0, "freq":   1, "cntLim":    0, "avgMsLim":      0, "avgKBLim":     0, "avgKBsLim":     0, "freqLim":   0, "freqLimSelf":   0, "troubleIndex":        0 }
      int timestampStartPos = s.indexOf("timestamp") + 13;
      int directionStartPos = s.indexOf("direction") + 13;
      String direction = s.substring(directionStartPos, directionStartPos + 3); // snd/rcv

      String month = s.substring(timestampStartPos, timestampStartPos + 7); // yyyy-MM
      String date = s.substring(timestampStartPos, timestampStartPos + 10); // yyyy-MM-dd

      String endpoint = s.substring(s.indexOf("endpoint") + 12, s.indexOf("endpoint") + 28); // 50V000000000113S
      String line = s.substring(s.indexOf("cnt")); // resten av linjen sendes inn i Summary.add()
      monthSummaryMap.merge(month, new OppositeSummary(month, direction, endpoint, delayLim, expireLim, line), (oldSummary, newSummary) -> oldSummary.add(endpoint, line));
      dateSummaryMap.merge(date, new OppositeSummary(date, direction, endpoint, delayLim, expireLim, line), (oldSummary, newSummary) -> oldSummary.add(endpoint, line));
    }
  }

}

